#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: GNU General Public License v2
#
# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2023-11-04
# File  : cisco_meraki_appliance_uplinks.py (check plugin)

# 2023-11-12: re-added appliance uplinks usage


from dataclasses import dataclass
from datetime import datetime
from _collections_abc import Mapping

from cmk.base.plugins.agent_based.agent_based_api.v1 import (
    register,
    Result,
    Service,
    State,
    Metric,
    render,
)
from cmk.base.plugins.agent_based.agent_based_api.v1.type_defs import (
    CheckResult,
    DiscoveryResult,
    StringTable,
)
from cmk.base.plugins.agent_based.utils.cisco_meraki import (
    load_json,
)


_LAST_REPORTED_AT = "%Y-%m-%dT%H:%M:%SZ"


@dataclass(frozen=True)
class ApplianceUplink:
    interface: str
    gateway: str
    ip: str
    ip_assigned_by: str
    primary_dns: str
    secondary_dns: str
    status: str
    public_ip: str
    sent: int | None
    received: int | None


@dataclass(frozen=True)
class ApplianceUplinkHA:
    enabled: bool
    role: str


@dataclass(frozen=True)
class Appliance:
    serial: str
    model: str
    last_reported_at: datetime
    network_name: str
    uplinks: Mapping[str, ApplianceUplink]
    high_availability: ApplianceUplinkHA


# sample string_table
__appliance_uplinks = [
    {
        'highAvailability': {
            'enabled': False,
            'role': 'primary'
        },
        'lastReportedAt': '2023-11-05T08:18:21Z',
        'model': 'MX68',
        'networkId': 'L_620933798623726672',
        'networkName': 'GS-Missen',
        'serial': 'Q2KY-L2KB-Z2YG',
        'uplinks': [
            {
                'gateway': '192.168.10.1',
                'interface': 'wan1',
                'ip': '192.168.10.2',
                'ipAssignedBy': 'static',
                'primaryDns': '192.168.10.1',
                'publicIp': '79.197.135.251',
                'secondaryDns': '9.9.9.9',
                'status': 'active',
                "received": 52320006,
                "sent": 52928038,
            },
            {
                'gateway': '192.168.5.100',
                'interface': 'wan2',
                'ip': '192.168.5.236',
                'ipAssignedBy': 'dhcp',
                'primaryDns': '192.168.5.100',
                'publicIp': '84.151.100.109',
                'secondaryDns': '0.0.0.0',
                'status': 'active'
            }
        ]
    }
]


def parse_appliance_uplinks(string_table: StringTable) -> Appliance | None:
    json_data = load_json(string_table)
    json_data = json_data[0]
    return Appliance(
        serial=json_data['serial'],
        model=json_data['model'],
        last_reported_at=datetime.strptime(json_data['lastReportedAt'], _LAST_REPORTED_AT),
        network_name=json_data['networkName'],
        high_availability=ApplianceUplinkHA(
            enabled=json_data['highAvailability']['enabled'],
            role=json_data['highAvailability']['role'],
        ),
        uplinks={
            uplink['interface']: ApplianceUplink(
                interface=uplink['interface'],
                gateway=uplink['gateway'],
                ip=uplink['ip'],
                ip_assigned_by=uplink['ipAssignedBy'],
                primary_dns=uplink['primaryDns'],
                public_ip=uplink['publicIp'],
                secondary_dns=uplink['secondaryDns'],
                status=uplink['status'],
                received=uplink.get('received', None),
                sent=uplink.get('sent', None)
            ) for uplink in json_data['uplinks']
        },
    )


register.agent_section(
    name="cisco_meraki_org_appliance_uplinks",
    parse_function=parse_appliance_uplinks,
)


def discover_appliance_uplinks(section: Appliance) -> DiscoveryResult:
    for uplink in section.uplinks.keys():
        yield Service(item=uplink)


_STATUS_MAP = {
    "active": 0,
    "ready": 0,
    "not connected": 1,
    "failed": 2,
}


def check_appliance_uplinks(item: str, params: Mapping[str, any], section: Appliance) -> CheckResult:
    try:
        uplink: ApplianceUplink = section.uplinks[item]
    except KeyError:
        return None

    if params.get('status_map'):
        _STATUS_MAP.update(params['status_map'])

    yield Result(state=State(_STATUS_MAP.get(uplink.status, 3)), summary=f'Status: {uplink.status}')
    yield Result(state=State.OK, summary=f'IP: {uplink.ip}')
    yield Result(state=State.OK, summary=f'Public IP: {uplink.public_ip}')
    yield Result(state=State.OK, notice=f'Network: {section.network_name}')

    if uplink.received:
        yield Result(state=State.OK, summary=f'in: {render.networkbandwidth(uplink.received)}')
        yield Metric(name='if_in_bps', value=uplink.received * 8)

    if uplink.sent:
        yield Result(state=State.OK, summary=f'Out: {render.networkbandwidth(uplink.sent)}')
        yield Metric(name='if_out_bps', value=uplink.sent * 8)

    # not needed, will show in device status
    # yield from check_last_reported_ts(last_reported_ts=section.last_reported_at.timestamp())

    # not sure if this is usefully, need system with H/A enabled=True to check
    yield Result(state=State.OK, notice=f'H/A enabled: {section.high_availability.enabled}')
    yield Result(state=State.OK, notice=f'H/A role: {section.high_availability.role}')

    yield Result(state=State.OK, notice=f'Gateway: {uplink.gateway}')
    yield Result(state=State.OK, notice=f'IP assigned by: {uplink.ip_assigned_by}')
    yield Result(state=State.OK, notice=f'Primary DNS: {uplink.primary_dns}')
    yield Result(state=State.OK, notice=f'Secondary DNS: {uplink.secondary_dns}')


register.check_plugin(
    name='cisco_meraki_org_appliance_uplinks',
    service_name='Appliance Uplink %s',
    discovery_function=discover_appliance_uplinks,
    check_function=check_appliance_uplinks,
    check_default_parameters={},
    check_ruleset_name='cisco_meraki_org_appliance_uplinks',
)
