#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: GNU General Public License v2
#
# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2023-11-04
# File  : cisco_meraki_appliance_vpns.py (check plugin)

from dataclasses import dataclass
from _collections_abc import Sequence, Mapping

from cmk.base.plugins.agent_based.agent_based_api.v1 import (
    register,
    Result,
    Service,
    State,
)
from cmk.base.plugins.agent_based.agent_based_api.v1.type_defs import (
    CheckResult,
    DiscoveryResult,
    StringTable,
)
from cmk.base.plugins.agent_based.utils.cisco_meraki import (
    load_json,
)


@dataclass(frozen=True)
class ApplianceVPNUplink:
    interface: str
    public_ip: str


@dataclass(frozen=True)
class ApplianceVPNPeer:
    type: str
    network_name: str
    reachability: str
    public_ip: str | None
    device_vpn_mode: str
    uplinks: Sequence[ApplianceVPNUplink]


# sample string_table
__appliance_vpn_statuses = [
    {
        'deviceSerial': 'Q2LY-666K-GM6Z',
        'deviceStatus': 'online',
        'exportedSubnets': [
            {
                'name': 'Management',
                'subnet': '172.18.10.0/24'
            },
            {
                'name': 'Clients',
                'subnet': '172.18.15.0/24'
            },
            {
                'name': 'Clients-WLAN',
                'subnet': '172.18.20.0/24'
            },
            {
                'name': 'Drucker',
                'subnet': '172.18.30.0/24'
            },
            {
                'name': 'VoIP-Schrankenanlage',
                'subnet': '172.18.35.0/24'
            }
        ],
        'merakiVpnPeers': [
            {
                'networkId': 'L_575897802350007343',
                'networkName': 'Park und Schloss',
                'reachability': 'reachable'
            },
            {
                'networkId': 'N_575897802350071601',
                'networkName': 'Bärenbrücker Str.4',
                'reachability': 'reachable'
            }
        ],
        'networkId': 'L_575897802350013029',
        'networkName': 'Schranken-Anlage',
        'serial': 'Q2LY-666K-GM6Z',
        'thirdPartyVpnPeers': [
            {
                'name': 'Site-to-Site-Azure',
                'publicIp': '20.113.170.100',
                'reachability': 'unreachable'
            }
        ],
        'uplinks': [
            {
                'interface': 'wan1',
                'publicIp': '80.187.86.103'
            }
        ],
        'vpnMode': 'hub'
    }
]


def parse_appliance_vpns(string_table: StringTable) -> Mapping[str, ApplianceVPNPeer]:
    json_data = load_json(string_table)
    json_data = json_data[0]

    vpn_uplinks = [ApplianceVPNUplink(
        interface=uplink['interface'],
        public_ip=uplink['publicIp'],
    ) for uplink in json_data['uplinks']]

    meraki_peers = {
        peer['networkName']: ApplianceVPNPeer(
            type='Meraki VPN Peer',
            network_name=peer['networkName'],
            reachability=peer['reachability'],
            public_ip=None,
            device_vpn_mode=json_data['vpnMode'],
            uplinks=vpn_uplinks,
        ) for peer in json_data.get('merakiVpnPeers', [])
    }
    third_party_peers = {
        peer['name']: ApplianceVPNPeer(
            type='third party VPN Peer',
            network_name=peer['name'],
            reachability=peer['reachability'],
            public_ip=peer['publicIp'],
            device_vpn_mode=json_data['vpnMode'],
            uplinks=vpn_uplinks,
        ) for peer in json_data.get('thirdPartyVpnPeers', [])
    }

    meraki_peers.update(third_party_peers)

    return meraki_peers


register.agent_section(
    name="cisco_meraki_org_appliance_vpns",
    parse_function=parse_appliance_vpns,
)


def discover_appliance_vpns(section: Mapping[str, ApplianceVPNPeer]) -> DiscoveryResult:
    for key in section:
        yield Service(item=key)


def check_appliance_vpns(item: str, params: Mapping[str, any], section: Mapping[str, ApplianceVPNPeer]) -> CheckResult:
    try:
        peer: ApplianceVPNPeer = section[item]
    except KeyError:
        return None

    if peer.reachability.lower() not in ['reachable']:
        yield Result(
            state=State(params.get('status_not_reachable', 1)),
            summary=f'{peer.reachability}',
        )
    else:
        yield Result(state=State.OK, summary=f'{peer.reachability}')

    yield Result(state=State.OK, summary=f'Type: {peer.type}')
    if peer.public_ip:
        yield Result(state=State.OK, summary=f'Peer IP: {peer.public_ip}')

    yield Result(state=State.OK, notice=f'VPN Mode: {peer.device_vpn_mode}')
    yield Result(state=State.OK, notice=f'Uplink(s):')
    for uplink in peer.uplinks:
        yield Result(state=State.OK, notice=f'name: {uplink.interface}, public IP: {uplink.public_ip}')


register.check_plugin(
    name='cisco_meraki_org_appliance_vpns',
    service_name='Appliance VPN %s',
    discovery_function=discover_appliance_vpns,
    check_function=check_appliance_vpns,
    check_default_parameters={},
    check_ruleset_name='cisco_meraki_org_appliance_vpns',
)
