#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: GNU General Public License v2
#
# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2023-11-13
# File  : cisco_meraki_org_cellular_uplinks.py (check plugin)


from dataclasses import dataclass
from datetime import datetime
from _collections_abc import Mapping

from cmk.base.plugins.agent_based.agent_based_api.v1 import (
    register,
    Result,
    Service,
    State,
    Metric,
    render,
)
from cmk.base.plugins.agent_based.agent_based_api.v1.type_defs import (
    CheckResult,
    DiscoveryResult,
    StringTable,
)
from cmk.base.plugins.agent_based.utils.cisco_meraki import (
    load_json,
)


_LAST_REPORTED_AT = "%Y-%m-%dT%H:%M:%SZ"


@dataclass(frozen=True)
class CellularUplink:
    apn: str
    connection_type: str
    dns1: str
    dns2: str
    gateway: str
    iccid: str
    interface: str
    ip: str
    model: str
    provider: str
    public_ip: str
    signal_type: str
    status: str
    rsrp: int
    rsrq: int
    sent: int | None
    received: int | None


@dataclass(frozen=True)
class CellularUplinkHA:
    enabled: bool
    role: str


@dataclass(frozen=True)
class CellularGateway:
    serial: str
    model: str
    last_reported_at: datetime
    # network_name: str
    uplinks: Mapping[str, CellularUplink]
    high_availability: CellularUplinkHA


__cellular_uplinks = [
    {
        "highAvailability": {
            "enabled": False,
            "role": "primary"
        },
        "lastReportedAt": "2023-11-13T19:52:06Z",
        "model": "MG41",
        "networkId": "L_575897802350007343",
        "serial": "Q2ZY-XLHP-Z8ND",
        "uplinks": [
            {
                "apn": "web.vodafone.de",
                "connectionType": "lte",
                "dns1": None,
                "dns2": None,
                "gateway": None,
                "iccid": "89492027206029876518",
                "interface": "cellular",
                "ip": None,
                "model": "integrated",
                "provider": "vodafone.de",
                "publicIp": "109.42.243.39",
                "signalStat": {
                    "rsrp": "-111",
                    "rsrq": "-8"
                },
                "signalType": None,
                "status": "active"
            }
        ]
    }
]


def parse_cellular_uplinks(string_table: StringTable) -> CellularGateway | None:
    json_data = load_json(string_table)
    json_data = json_data[0]
    return CellularGateway(
        serial=json_data['serial'],
        model=json_data['model'],
        last_reported_at=datetime.strptime(json_data['lastReportedAt'], _LAST_REPORTED_AT),
        # network_name=json_data['networkName'],
        high_availability=CellularUplinkHA(
            enabled=json_data['highAvailability']['enabled'],
            role=json_data['highAvailability']['role'],
        ),
        uplinks={
            uplink['interface']: CellularUplink(
                apn=uplink['apn'],
                connection_type=uplink['connectionType'],
                dns1=uplink['dns1'],
                dns2=uplink['dns2'],
                gateway=uplink['gateway'],
                iccid=uplink['iccid'],
                interface=uplink['interface'],
                ip=uplink['ip'],
                model=uplink['model'],
                provider=uplink['provider'],
                public_ip=uplink['publicIp'],
                signal_type=uplink['signalType'],
                status=uplink['status'],
                rsrp=uplink['signalStat']['rsrp'],
                rsrq=uplink['signalStat']['rsrq'],
                received=uplink.get('received', None),
                sent=uplink.get('sent', None),
            ) for uplink in json_data['uplinks']
        },
    )


register.agent_section(
    name="cisco_meraki_org_cellular_uplinks",
    parse_function=parse_cellular_uplinks,
)


def discover_cellular_uplinks(section: CellularGateway) -> DiscoveryResult:
    for uplink in section.uplinks.keys():
        yield Service(item=uplink)


def check_cellular_uplinks(item: str, params: Mapping[str, any], section: CellularGateway) -> CheckResult:
    try:
        uplink: CellularUplink = section.uplinks[item]
    except KeyError:
        return None

    if uplink.status not in ['active']:
        yield Result(state=State(params.get('status_not_active', 1)), summary=f'Status: {uplink.status}')
    else:
        yield Result(state=State.OK, summary=f'Status: {uplink.status}')
    yield Result(state=State.OK, notice=f'IP: {uplink.ip}')
    yield Result(state=State.OK, notice=f'APN: {uplink.apn}')
    yield Result(state=State.OK, notice=f'Provider: {uplink.provider}')
    yield Result(state=State.OK, summary=f'Public IP: {uplink.public_ip}')
    yield Result(state=State.OK, summary=f'Connection type: {uplink.connection_type}')
    yield Result(state=State.OK, notice=f'ICCID: {uplink.iccid}')
    yield Result(state=State.OK, notice=f'Signal type: {uplink.signal_type}')

    # yield Result(state=State.OK, notice=f'Network: {section.network_name}')

    if uplink.rsrp:
        yield Result(state=State.OK, summary=f'RSRP: {uplink.rsrp} dBm')

    if uplink.rsrp:
        yield Result(state=State.OK, summary=f'RSRQ: {uplink.rsrq} dB')

    if uplink.received:
        yield Result(state=State.OK, summary=f'in: {render.networkbandwidth(uplink.received)}')
        yield Metric(name='if_in_bps', value=uplink.received * 8)

    if uplink.sent:
        yield Result(state=State.OK, summary=f'Out: {render.networkbandwidth(uplink.sent)}')
        yield Metric(name='if_out_bps', value=uplink.sent * 8)

    # not needed, will show in device status
    # yield from check_last_reported_ts(last_reported_ts=section.last_reported_at.timestamp())

    # not sure if this is usefully, need system with H/A enabled=True to check
    yield Result(state=State.OK, notice=f'H/A enabled: {section.high_availability.enabled}')
    yield Result(state=State.OK, notice=f'H/A role: {section.high_availability.role}')

    yield Result(state=State.OK, notice=f'Gateway: {uplink.gateway}')
    # yield Result(state=State.OK, notice=f'IP assigned by: {uplink.ip_assigned_by}')
    yield Result(state=State.OK, notice=f'DNS 1: {uplink.dns1}')
    yield Result(state=State.OK, notice=f'DNS 2: {uplink.dns2}')


register.check_plugin(
    name='cisco_meraki_org_cellular_uplinks',
    service_name='Cellular Uplink %s',
    discovery_function=discover_cellular_uplinks,
    check_function=check_cellular_uplinks,
    check_default_parameters={},
    check_ruleset_name='cisco_meraki_org_cellular_uplinks',
)
