#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: GNU General Public License v2
#
# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2023-11-11
# File  : cisco_meraki_org_wireless_device_status.py (check plugin)

from dataclasses import dataclass
from _collections_abc import Mapping, Sequence

from cmk.base.plugins.agent_based.agent_based_api.v1 import (
    register,
    Result,
    Service,
    State,
    Metric,
)
from cmk.base.plugins.agent_based.agent_based_api.v1.type_defs import (
    CheckResult,
    DiscoveryResult,
    StringTable,
)
from cmk.base.plugins.agent_based.utils.cisco_meraki import (
    load_json,
)


@dataclass(frozen=True)
class SSDI:
    name: str
    number: int
    band: str
    channel: int
    channel_width: str
    power: str
    broadcasting: bool
    bssid: str
    enabled: bool
    visible: bool


def parse_wireless_device_status(string_table: StringTable):
    json_data = load_json(string_table)
    json_data = json_data[0]

    ssids = {}
    for row in json_data['basicServiceSets']:
        if row['ssidName'].startswith('Unconfigured SSID'):
            continue  # ignore unconfigured entry's

        item = str(row['ssidNumber']) + ' on band ' + row['band']

        ssids[item] = SSDI(
            name=row['ssidName'],
            number=row['ssidNumber'],
            band=row['band'],
            channel=row['channel'],
            channel_width=row['channelWidth'],
            power=row['power'],
            broadcasting=row['broadcasting'],
            bssid=row['bssid'],
            enabled=row['enabled'],
            visible=row['visible'],
        )

    return ssids


register.agent_section(
    name="cisco_meraki_org_wireless_device_status",
    parse_function=parse_wireless_device_status,
)

_is = {
    True: '',
    False: ' not',
}


def discover_wireless_device_status(section: Mapping[str, SSDI]) -> DiscoveryResult:
    for ssid in section.keys():
        yield Service(item=ssid)


def check_wireless_device_status(
        item: str,
        params: Mapping[str, any],
        section: Mapping[str, SSDI]
) -> CheckResult:
    try:
        ssid: SSDI = section[item]
    except KeyError:
        return None

    yield Result(state=State.OK, summary=f'Name: {ssid.name}')
    if not ssid.enabled:
        yield Result(state=State(params['state_if_not_enabled']), summary=f'is{_is[ssid.enabled]} enabled')
    else:
        yield Result(state=State.OK, summary=f'is{_is[ssid.enabled]} enabled')

        # don't show details if SSID not enabled
        yield Result(state=State.OK, notice=f'is{_is[ssid.visible]} visible')
        yield Result(state=State.OK, notice=f'is{_is[ssid.broadcasting]} broadcasting')
        yield Result(state=State.OK, notice=f'BSSID: {ssid.bssid}')
        yield Result(state=State.OK, notice=f'Band: {ssid.band}')
        yield Result(state=State.OK, summary=f'Channel: {ssid.channel}')
        yield Metric(name='channel', value=ssid.channel)
        yield Result(state=State.OK, summary=f'Channel width: {ssid.channel_width}')
        yield Metric(name='channel_width', value=int(ssid.channel_width.split(' ')[0]) * 1000000)  # change to Hz from MHz
        yield Result(state=State.OK, summary=f'Power: {ssid.power}')
        yield Metric(name='signal_power', value=int(ssid.power.split(' ')[0]))


register.check_plugin(
    name='cisco_meraki_org_wireless_device_status',
    service_name='SSID %s',
    discovery_function=discover_wireless_device_status,
    check_function=check_wireless_device_status,
    check_default_parameters={
        'state_if_not_enabled': 1,
    },
    check_ruleset_name='cisco_meraki_wireless_device_status',
)

