#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: GNU General Public License v2
#
# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2023-11-11
# File  : cisco_meraki_organisations.py (check plugin)

from dataclasses import dataclass
from _collections_abc import Mapping

from cmk.base.plugins.agent_based.agent_based_api.v1 import (
    register,
    TableRow,
    Service,
    State,
    Result,
)
from cmk.base.plugins.agent_based.agent_based_api.v1.type_defs import (
    StringTable,
    InventoryResult,
    DiscoveryResult,
    CheckResult,
)
from cmk.base.plugins.agent_based.utils.cisco_meraki import (
    load_json,
)

__orgaisations = [
    {'id': '670771',
     'name': 'Bechtle GmbH & Co KG',
     'url': 'https://n229.meraki.com/o/lt-Wka/manage/organization/overview',
     'api': {'enabled': True},
     'licensing': {'model': 'co-term'},
     'cloud': {'region': {'name': 'Europe'}},
     'management': {'details': []}},
]


@dataclass(frozen=True)
class Organisation:
    id: str
    name: str
    url: str
    api: bool
    licensing: str
    cloud: bool
    management: Mapping[any] | None


_api_status = {
    True: 'enabled',
    False: 'disabled',
}


def parse_meraki_organisations(string_table: StringTable) -> Mapping[str, Organisation] | None:
    json_data = load_json(string_table)
    return {
        org['id']: Organisation(
            id=org['id'],
            name=org['name'],
            url=org['url'],
            api=org['api']['enabled'],
            licensing=org['licensing']['model'],
            cloud=org['cloud']['region']['name'],
            management=org['management'],
        ) for org in json_data
    }


register.agent_section(
    name='cisco_meraki_org_organisations',
    parsed_section_name='cisco_meraki_organisations_api',
    parse_function=parse_meraki_organisations,
)


def inventory_meraki_organisations(section: Mapping[str, Organisation]) -> InventoryResult:
    path = ['software', 'applications', 'cisco_meraki', 'organisations']

    for org in section.values():
        yield TableRow(
            path=path,
            key_columns={'org_id': org.id},
            inventory_columns={
                'org_name': org.name,
                'url': org.url,
                'api': _api_status[org.api],
                'licensing': org.licensing,
                'cloud': org.cloud,
            }
        )


register.inventory_plugin(
    name='cisco_meraki_organisations_api',
    # sections=['cisco_meraki_organisations'],
    inventory_function=inventory_meraki_organisations,
)


def discover_organisations_api(params: Mapping[str, str], section: Mapping[str, Organisation]) -> DiscoveryResult:
    for organisation_id in section:
        match params['item_variant']:
            case 'org_id':
                item = organisation_id
            case 'org_name':
                item = f'{section[organisation_id].name}'
            case _:
                item = f'{section[organisation_id].name}/{organisation_id}'

        yield Service(
            item=item,
            parameters={
                'internal_item_name': organisation_id,
                'item_variant': params['item_variant'],
                # 'old_item_name': f'{section[organisation_id].organisation_name}/{organisation_id}',
            })


def check_organisations_api(
    item: str,
    params: Mapping[str, any],
    section: Mapping[str, Organisation],
) -> CheckResult:
    if (org := section.get(params.get('internal_item_name'))) is None:
        return

    yield Result(
        state=State.OK if org.api else State(params['state_api_not_enabled']),
        summary=f'Status: {_api_status[org.api]}',
    )

    org_id = f'ID: {org.id}'
    org_name = f'Name: {org.name}'
    org_id_notice = f'Organisation ID: {org.id}'
    org_name_notice = f'Organisation name: {org.name}'

    match params['item_variant']:
        case 'org_id':
            yield Result(state=State.OK, notice=org_id_notice)
            if params.get('dont_show_alias_on_info'):
                yield Result(state=State.OK, notice=org_name_notice)
            else:
                yield Result(state=State.OK, summary=org_name)
        case 'org_name':
            if params.get('dont_show_alias_on_info'):
                yield Result(state=State.OK, notice=org_id_notice)
            else:
                yield Result(state=State.OK, summary=org_id)
            yield Result(state=State.OK, notice=org_name_notice)

        case _:
            yield Result(state=State.OK, notice=org_id_notice)
            yield Result(state=State.OK, notice=org_name_notice)




register.check_plugin(
    name='cisco_meraki_organisations_api',
    service_name='Cisco Meraki API %s',
    discovery_function=discover_organisations_api,
    discovery_ruleset_name='discovery_meraki_organisations',
    discovery_default_parameters={
        'item_variant': 'org_id_name',
    },
    check_function=check_organisations_api,
    check_ruleset_name='cisco_meraki_organisations_api',
    check_default_parameters={
        'state_api_not_enabled': 1,
    },
)