#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: GNU General Public License v2
#
# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2023-11-11
# File  : cisco_meraki_switch_ports_statuses.py (check plugin)

from dataclasses import dataclass
from _collections_abc import Mapping, Sequence

from cmk.base.plugins.agent_based.agent_based_api.v1 import (
    register,
    Result,
    Service,
    State,
    TableRow,
)
from cmk.base.plugins.agent_based.agent_based_api.v1.type_defs import (
    CheckResult,
    DiscoveryResult,
    StringTable,
    InventoryResult,
)
from cmk.base.plugins.agent_based.utils.cisco_meraki import (
    load_json,
)


@dataclass(frozen=True)
class SwitchSecurePort:
    active: bool
    authentication_status: str
    configOverrides: Mapping[any]
    enabled: bool


@dataclass(frozen=True)
class SwitchPortCDP:
    device_id: str
    device_port: str
    address: str
    capabilities: str
    platform: str
    version: str


@dataclass(frozen=True)
class SwitchPortLLDP:
    chassis_id: str
    port_id: str
    system_name: str | None
    system_description: str | None
    port_description: str | None
    cache_capabilities: str | None
    management_address: str | None


@dataclass(frozen=True)
class SwitchPort:
    port_id: str
    client_count: int
    duplex: str
    enabled: bool
    errors: Sequence[str]
    is_up_link: bool
    power_usage_in_wh: float | None
    speed: str
    status: str
    warnings: Sequence[str]
    secure_port: SwitchSecurePort | None
    lldp: Mapping[str, any] | None
    cdp: SwitchPortCDP | None


_admin_status = {
    True: 'enabled',
    False: 'disabled',
}

_is_up_link = {
    True: 'yes',
    False: 'no',
}


def parse_switch_ports_statuses(string_table: StringTable) -> Mapping[str, SwitchPort] | None:
    json_data = load_json(string_table)
    json_data = json_data[0]

    def _parse_cdp(cdp: Mapping[str, str] | None) -> SwitchPortCDP | None:
        return SwitchPortCDP(
            device_id=cdp['deviceId'],
            device_port=cdp['portId'],
            address=cdp['address'],
            capabilities=cdp['capabilities'],
            platform=cdp['platform'],
            version=cdp['version'],
        ) if cdp else None

    def _parse_lldp(lldp: Mapping[str, str] | None) -> SwitchPortLLDP | None:
        return SwitchPortLLDP(
            chassis_id=lldp['chassisId'],
            port_id=lldp['portId'],
            system_name=lldp.get('systemName', None),
            system_description=lldp.get('systemDescription', None),
            port_description=lldp.get('portDescription', None),
            cache_capabilities=lldp.get('systemCapabilities', None),
            management_address=lldp.get('managementAddress', None)
        ) if lldp else None

    def _parse_secure_port(secure_port: Mapping[str, any] | None) -> SwitchSecurePort | None:
        return SwitchSecurePort(
            active=secure_port['active'],
            authentication_status=secure_port['authenticationStatus'],
            configOverrides=secure_port['configOverrides'],
            enabled=secure_port['enabled']
        ) if secure_port else None

    return {
        port['portId']: SwitchPort(
            port_id=port['portId'],
            client_count=port['clientCount'],
            duplex=port['duplex'],
            enabled=port['enabled'],
            errors=port['errors'],
            is_up_link=port['isUplink'],
            power_usage_in_wh=port.get('powerUsageInWh', None),
            speed=port['speed'],
            status=port['status'],
            warnings=port['warnings'],
            lldp=_parse_lldp(port.get('lldp', None)),
            cdp=_parse_cdp(port.get('cdp', None)),
            secure_port=_parse_secure_port(port.get('securePort', None)),
        ) for port in json_data
    }


register.agent_section(
    name="cisco_meraki_org_switch_ports_statuses",
    parse_function=parse_switch_ports_statuses,
)


def discover_switch_ports_statuses(section: Mapping[str, SwitchPort]) -> DiscoveryResult:
    for port in section.keys():
        yield Service(
            item=port,
            parameters={
                'enabled': section[port].enabled,
                'status': section[port].status,
                'speed': section[port].speed,
            }
        )


def check_switch_ports_statuses(item: str, params: Mapping[str, any], section: Mapping[str, SwitchPort]) -> CheckResult:
    def _status_changed(is_state: str, was_state: str, state: int, message: str):
        if is_state != was_state:
            is_state = is_state if is_state else 'N/A'
            was_state = was_state if was_state else 'N/A'
            yield Result(state=State(state), notice=f'{message}: from {was_state}, to {is_state}')

    try:
        port: SwitchPort = section[item]
    except KeyError:
        return

    # check admin state changed
    yield from _status_changed(
        is_state=_admin_status[port.enabled],
        was_state=_admin_status[params['enabled']],
        message='Admin status changed',
        state=params['state_admin_change'],
    )

    if port.enabled:  # check admin sate
        yield Result(state=State.OK, notice=f'Admin status: {_admin_status[port.enabled]}')
        # check operational status changed
        yield from _status_changed(
            is_state=port.status.lower(),
            was_state=params['status'].lower(),
            message='Operational status changed',
            state=params['state_op_change']
        )
        if port.status.lower() == 'connected':  # check operational state
            yield Result(state=State.OK, notice=f'Operational status: {port.status}')
            # check speed changed
            yield from _status_changed(
                is_state=port.speed,
                was_state=params['speed'],
                message='Speed changed',
                state=params['state_speed_change']
            )
            if params['speed'] == port.speed:  # only if speed unchanged
                yield Result(state=State.OK, summary=f'Speed: {port.speed}')
            if port.duplex.lower() == 'full':  # check duplex state
                yield Result(state=State.OK, summary=f'Duplex: {port.duplex}')
            else:
                yield Result(state=State(params['state_not_full_duplex']), summary=f'Duplex: {port.duplex}')
            yield Result(state=State.OK, summary=f'Clients: {port.client_count}')
        else:
            yield Result(state=State(params['state_not_connected']), summary=f'Operational status: {port.status}')
    else:
        yield Result(state=State(params['state_disabled']), summary=f'Admin status: {_admin_status[port.enabled]}')

    if port.is_up_link:
        yield Result(state=State.OK, summary=f'UP-Link: {_is_up_link[port.is_up_link]}')
    else:
        yield Result(state=State.OK, notice=f'UP-Link: {_is_up_link[port.is_up_link]}')

    if port.power_usage_in_wh:
        yield Result(state=State.OK, summary=f'Power usage: {port.power_usage_in_wh} Wh')

    for warning in port.warnings:
        yield Result(state=State.WARN, notice=f'{warning}')
    for error in port.errors:
        if error not in ['Port disconnected', 'Port disabled']:
            yield Result(state=State.CRIT, notice=f'{error}')

    if port.secure_port and port.secure_port.enabled:
        yield Result(state=State.UNKNOWN, summary=f'Secure Port enabled', details=f'Secure Port: {port.secure_port}')


register.check_plugin(
    name='cisco_meraki_org_switch_ports_statuses',
    service_name='Port %s',
    discovery_function=discover_switch_ports_statuses,
    check_function=check_switch_ports_statuses,
    check_default_parameters={
        'state_disabled': 0,
        'state_not_connected': 0,
        'state_not_full_duplex': 1,
        'state_speed_change': 1,
        'state_admin_change': 1,
        'state_op_change': 1,
    },
    check_ruleset_name='cisco_meraki_switch_ports_statuses',
)


def inventory_meraki_cdp_cache(section: Mapping[str, SwitchPort]) -> InventoryResult:
    path = ['networking', 'cdp_cache', 'neighbours']

    for port in section.values():
        if cdp := port.cdp:
            key_columns = {'local_port': port.port_id}
            neighbour = {
                'neighbour_name': cdp.device_id,
                'neighbour_port': cdp.device_port,
                'neighbour_address': cdp.address,
                'platform_details': cdp.version,
                'platform': cdp.platform,
                'capabilities': cdp.capabilities
            }
            yield TableRow(
                path=path,
                key_columns=key_columns,
                inventory_columns=neighbour
            )


register.inventory_plugin(
    name='inv_meraki_cdp_cache',
    sections=['cisco_meraki_org_switch_ports_statuses'],
    inventory_function=inventory_meraki_cdp_cache,
)


def inventory_meraki_lldp_cache(section: Mapping[str, SwitchPort]) -> InventoryResult:
    path = ['networking', 'lldp_cache', 'neighbours']

    for port in section.values():
        if lldp := port.lldp:
            key_columns = {'local_port': port.port_id}
            neighbour = {
                'neighbour_id': lldp.chassis_id,
                'neighbour_port': lldp.port_id,
                'neighbour_name': lldp.system_name,
                'system_description': lldp.system_description,
                'port_description': lldp.port_description,
                'capabilities': lldp.cache_capabilities,
                'neighbour_address': lldp.management_address,
            }
            yield TableRow(
                path=path,
                key_columns=key_columns,
                inventory_columns=neighbour
            )


register.inventory_plugin(
    name='inv_meraki_lldp_cache',
    sections=['cisco_meraki_org_switch_ports_statuses'],
    inventory_function=inventory_meraki_lldp_cache,
)
