#!/usr/bin/env python3
# Copyright (C) 2022 Checkmk GmbH - License: GNU General Public License v2
# This file is part of Checkmk (https://checkmk.com). It is subject to the terms and
# conditions defined in the file COPYING, which is part of this source code package.

# enhancements by thl-cmk[at]outlook[dot]com, https://thl-cmk.hopto.org
# - moved section name to utils/cisco_meraki.py (reuse with WATO)
# - added option to adjust piggyback names with prefix, suffix, upper/lower case
# - added section _SEC_NAME_DEVICE_UPLINKS_INFO (inventory of uplink information)
# - added section _SEC_NAME_APPLIANCE_UPLINKS (status and usage of appliance uplinks)
# - updated Cisco meraki dashboard SDK from version 1.27.0 to 1.38.0 (needed for appliance uplinks)
# - added section _SEC_NAME_APPLIANCE_VPNS (status and usage of appliance uplinks)
#
# 2023-11-09: removed nonworking original datacache (cleanup before reimplementing)
#             added new Cacheing code based on AWS agent - works this time ;-)
# 2023-11-10: a bit of cleanup, removed oneliner API calls, moved to the calling function
#             moved add networks to MerakiOrganisation.query
#             added cisco_meraki_org_networks output section (base to create networks inventory)
# 2023-10-11: fixed _get_device_piggyback drop hosts without names -> can be changed to use serial instead
#             fixed in device info drop hosts without names -> can be changed to use serial instead
#             fixed do nit access organisations where the API is not enabled
#             added organisation overview section _SEC_NAME_ORGANISATIONS to discover orgs with API not enabled
# 2023-11-12: added --no-cache option
#             re-added appliance uplink usage
# 2023-11-13: optimized API request: don't ask for device information's if there are no devices in an organisation
#             API requests by product type: don't ask for products that are not available in the organisation
# 2023-11-17: changed to always get organisations instead of organisation_by_id
# 2023-11-18: changed from include sections to exclude sections
# 2023-11-19: changed if no hostname (piggyback) to use serial+device type
# 2024-04-05: fixed min SDK version for MerakiGetOrganizationApplianceUplinksUsageByNetwork and
#                                       MerakiGetOrganizationWirelessDevicesEthernetStatuses

# ToDo: create inventory from Networks, is per organisation, not sure where/how to put this in the inventory
# ToDo: list Connected Datacenters like Umbrella https://developer.cisco.com/meraki/api-v1/list-data-centers/

# ToDo: if the following is available (right now only with early access enabled)
# ToDO: https://developer.cisco.com/meraki/api-v1/get-organization-switch-ports-statuses-by-switch/
# ToDo: https://developer.cisco.com/meraki/api-v1/get-organization-switch-ports-overview/
#
# Meraki SDK on GitHub: https://github.com/meraki/dashboard-api-python
#

from __future__ import annotations

import argparse
import logging
from collections.abc import Iterable, Iterator, Mapping, Sequence
from dataclasses import dataclass
from enum import auto, Enum
from pathlib import Path
from typing import Final, TypedDict, Any
from time import time_ns
from random import randrange

import meraki  # type: ignore[import]

from cmk.utils.paths import tmp_dir

from cmk.special_agents.utils.agent_common import (
    ConditionalPiggybackSection,
    SectionWriter,
    special_agent_main,
)
from cmk.special_agents.utils.argument_parsing import Args, create_default_argument_parser
from cmk.special_agents.utils.misc import DataCache

from cmk.base.plugins.agent_based.utils.cisco_meraki import (
    MerakiNetwork,
    _SEC_NAME_APPLIANCE_UPLINKS,
    _SEC_NAME_APPLIANCE_VPNS,
    _SEC_NAME_CELLULAR_UPLINKS,
    _SEC_NAME_DEVICE_INFO,
    _SEC_NAME_DEVICE_STATUSES,
    _SEC_NAME_DEVICE_UPLINKS_INFO,
    _SEC_NAME_LICENSES_OVERVIEW,
    _SEC_NAME_NETWORKS,
    _SEC_NAME_ORGANISATIONS,
    _SEC_NAME_SENSOR_READINGS,
    _SEC_NAME_SWITCH_PORTS_STATUSES,
    _SEC_NAME_WIRELESS_DEVICE_STATUS,
    _SEC_NAME_WIRELESS_ETHERNET_STATUSES,
)

_LOGGER = logging.getLogger("agent_cisco_meraki")

_API_NAME_API: Final = "api"
_API_NAME_DEVICE_NAME: Final = "name"
_API_NAME_DEVICE_PRODUCT_TYPE: Final = 'productType'
_API_NAME_DEVICE_SERIAL: Final = "serial"
_API_NAME_DEVICE_TYPE_APPLIANCE: Final = 'appliance'
_API_NAME_DEVICE_TYPE_CAMERA: Final = 'camera'
_API_NAME_DEVICE_TYPE_CELLULAR: Final = 'cellularGateway'
_API_NAME_DEVICE_TYPE_SENSOR: Final = 'sensor'
_API_NAME_DEVICE_TYPE_SWITCH: Final = 'switch'
_API_NAME_DEVICE_TYPE_WIRELESS: Final = 'wireless'
_API_NAME_ENABLED: Final = "enabled"
_API_NAME_NETWORK_ID: Final = 'networkId'
_API_NAME_ORGANISATION_ID: Final = "id"
_API_NAME_ORGANISATION_NAME: Final = "name"

_SECTION_NAME_MAP = {
    _SEC_NAME_APPLIANCE_UPLINKS: "appliance_uplinks",
    _SEC_NAME_APPLIANCE_VPNS: "appliance_vpns",
    _SEC_NAME_CELLULAR_UPLINKS: "cellular_uplinks",
    _SEC_NAME_DEVICE_INFO: "device_info",
    _SEC_NAME_DEVICE_STATUSES: "device_status",
    _SEC_NAME_DEVICE_UPLINKS_INFO: "device_uplinks_info",
    _SEC_NAME_LICENSES_OVERVIEW: "licenses_overview",
    _SEC_NAME_NETWORKS: "networks",
    _SEC_NAME_ORGANISATIONS: "organisations",
    _SEC_NAME_SENSOR_READINGS: "sensor_readings",
    _SEC_NAME_SWITCH_PORTS_STATUSES: "switch_ports_statuses",
    _SEC_NAME_WIRELESS_DEVICE_STATUS: "wireless_device_status",
    _SEC_NAME_WIRELESS_ETHERNET_STATUSES: "wireless_ethernet_statuses",
}

_MIN_CACHE_INTERVAL = 300
_RANDOM_CACHE_INTERVAL = 300

MerakiCacheFilePath = Path(tmp_dir) / "agents" / "agent_cisco_meraki"
MerakiAPIData = Mapping[str, object]


#   .--dashboard-----------------------------------------------------------.
#   |              _           _     _                         _           |
#   |           __| | __ _ ___| |__ | |__   ___   __ _ _ __ __| |          |
#   |          / _` |/ _` / __| '_ \| '_ \ / _ \ / _` | '__/ _` |          |
#   |         | (_| | (_| \__ \ | | | |_) | (_) | (_| | | | (_| |          |
#   |          \__,_|\__,_|___/_| |_|_.__/ \___/ \__,_|_|  \__,_|          |
#   |                                                                      |
#   '----------------------------------------------------------------------'


def _configure_meraki_dashboard(
    api_key: str,
    debug: bool,
    proxy: str | None,
) -> meraki.DashboardAPI:
    return meraki.DashboardAPI(
        api_key=api_key,
        print_console=True,
        output_log=False,
        suppress_logging=not debug,
        requests_proxy=proxy,
    )


# .
#   .--section-------------------------------------------------------------.
#   |                                 _   _                                |
#   |                   ___  ___  ___| |_(_) ___  _ __                     |
#   |                  / __|/ _ \/ __| __| |/ _ \| '_ \                    |
#   |                  \__ \  __/ (__| |_| | (_) | | | |                   |
#   |                  |___/\___|\___|\__|_|\___/|_| |_|                   |
#   |                                                                      |
#   '----------------------------------------------------------------------'

@dataclass(frozen=True)
class MerakiConfig:
    dashboard: meraki.DashboardAPI
    hostname: str
    # section_names: Sequence[str]
    excluded_sections: Sequence[str]
    use_cache: bool
    org_id_as_prefix: bool


class MerakiAPIDataSource(Enum):
    org = auto()


@dataclass(frozen=True)
class Section:
    api_data_source: MerakiAPIDataSource
    name: str
    data: MerakiAPIData
    piggyback: str | None = None

    def get_name(self) -> str:
        return "_".join(["cisco_meraki", self.api_data_source.name, self.name])


class _Organisation(TypedDict):
    # See https://developer.cisco.com/meraki/api-v1/#!get-organizations
    # get the latest SDK: https://github.com/meraki/dashboard-api-python
    # if you want to extend this
    id: str
    name: str


# .
#   .--caches--------------------------------------------------------------.
#   |                                  _                                   |
#   |                    ___ __ _  ___| |__   ___  ___                     |
#   |                   / __/ _` |/ __| '_ \ / _ \/ __|                    |
#   |                  | (_| (_| | (__| | | |  __/\__ \                    |
#   |                   \___\__,_|\___|_| |_|\___||___/                    |
#   |                                                                      |
#   '----------------------------------------------------------------------'
#
# --\ DataCache
#   |
#   |--\ MerakiSection
#      |  - adds cache_interval = 86400
#      |  - adds get_validity_from_args = True
#      |
#      |--> MerakiGetOrganizations                          		-> default 86400
#      |
#      |--\ MerakiSectionOrg
#      |  |  - adds org_id parameter
#      |  |
#      |  |--> MerakiGetOrganization				                -> default 86400
#      |  |--> MerakiGetOrganizationLicensesOverview		        -> default 86400
#      |  |--> MerakiGetOrganizationDevices				            -> default 86400
#      |  |--> MerakiGetOrganizationNetworks			            -> default 86400
#      |  |--> MerakiGetOrganizationDevicesStatuses			        -> ex. 60+
#      |  |--> MerakiGetOrganizationDevicesUplinksAddressesByDevice	-> ex. 60+
#      |  |--> MerakiGetOrganizationApplianceUplinkStatuses		    -> ex. 60+
#      |  |--> MerakiGetOrganizationApplianceUplinksUsageByNetwork	-> ex. 59
#      |  |--> MerakiGetOrganizationApplianceVpnStatuses		    -> ex. 60+
#      |  |--> MerakiGetOrganizationSensorReadingsLatest		    -> ex. 60+
#      |  |--> MerakiGetOrganizationWirelessDevicesEthernetStatuses	-> ex. 60+
#      |  |--> MerakiGetOrganizationCellularGatewayUplinkStatuses   -> ex. 60.
#      |
#      |--\ MerakiSectionSerial
#         |  - adds serial as parameter
#         |  - sets cache_interval = 60 + randrange(300)
#         |
#         |--> MerakiGetDeviceSwitchPortsStatuses 	                -> default 60+
#         |--> MerakiGetDeviceWirelessStatus	                    -> default 60+

class MerakiSection(DataCache):
    def __init__(
            self,
            config: MerakiConfig,
            cache_interval: int = 86400,
    ):
        self._config = config
        self._received_results = {}
        self._cache_dir = MerakiCacheFilePath / self._config.hostname
        self._cache_file = MerakiCacheFilePath / self._config.hostname / self.name
        self._cache_interval = cache_interval
        super().__init__(self._cache_dir, self.name)

    @property
    def cache_interval(self):
        return self._cache_interval

    def get_validity_from_args(self, *args: Any) -> bool:
        # always True, for now there are no changing arguments, related to the cache
        return True


class MerakiSectionOrg(MerakiSection):
    def __init__(
            self,
            config: MerakiConfig,
            org_id: str,
            cache_interval: int = 86400,
    ):
        self._org_id = org_id
        super().__init__(config=config, cache_interval=cache_interval)


class MerakiSectionSerial(MerakiSection):
    def __init__(
            self,
            config: MerakiConfig,
            serial: str,
            cache_interval: int = _MIN_CACHE_INTERVAL + randrange(_RANDOM_CACHE_INTERVAL),
    ):
        self._serial = serial
        super().__init__(config=config, cache_interval=cache_interval)


class MerakiGetOrganizations(MerakiSection):
    @property
    def name(self):
        return "getOrganizations"

    def get_live_data(self):
        try:
            return self._config.dashboard.organizations.getOrganizations()
        except meraki.exceptions.APIError as e:
            _LOGGER.debug("Get organisations: %r", e)
            return []


class MerakiGetOrganization(MerakiSectionOrg):
    @property
    def name(self):
        return f'getOrganization_{self._org_id}'

    def get_live_data(self):
        try:
            return self._config.dashboard.organizations.getOrganization(self._org_id)
        except meraki.exceptions.APIError as e:
            _LOGGER.debug(f'Get organisation by id {self._org_id}: {e}')
            return {}


class MerakiGetOrganizationLicensesOverview(MerakiSectionOrg):
    @property
    def name(self):
        return f'getOrganizationLicensesOverview_{self._org_id}'

    def get_live_data(self):
        try:
            return self._config.dashboard.organizations.getOrganizationLicensesOverview(self._org_id)
        except meraki.exceptions.APIError as e:
            _LOGGER.debug("Organisation ID: %r: Get license overview: %r", self._org_id, e)
            return []


class MerakiGetOrganizationDevices(MerakiSectionOrg):
    @property
    def name(self):
        return f'getOrganizationDevices_{self._org_id}'

    def get_live_data(self):
        try:
            return self._config.dashboard.organizations.getOrganizationDevices(self._org_id, total_pages="all")
        except meraki.exceptions.APIError as e:
            _LOGGER.debug("Organisation ID: %r: Get devices: %r", self._org_id, e)
            return {}


class MerakiGetOrganizationNetworks(MerakiSectionOrg):
    @property
    def name(self):
        return f'getOrganizationNetworks_{self._org_id}'

    def get_live_data(self):
        try:
            return self._config.dashboard.organizations.getOrganizationNetworks(self._org_id, total_pages="all")
        except meraki.exceptions.APIError as e:
            _LOGGER.debug("Organisation ID: %r: Get networks: %r", self._org_id, e)
            return []


class MerakiGetOrganizationDevicesStatuses(MerakiSectionOrg):
    @property
    def name(self):
        return f'getOrganizationDevicesStatuses_{self._org_id}'

    def get_live_data(self):
        try:
            return self._config.dashboard.organizations.getOrganizationDevicesStatuses(self._org_id, total_pages="all")
        except meraki.exceptions.APIError as e:
            _LOGGER.debug("Organisation ID: %r: Get device statuses: %r", self._org_id, e)
            return []


class MerakiGetOrganizationDevicesUplinksAddressesByDevice(MerakiSectionOrg):
    @property
    def name(self):
        return f'getOrganizationDevicesUplinksAddressesByDevice_{self._org_id}'

    def get_live_data(self):
        try:
            return self._config.dashboard.organizations.getOrganizationDevicesUplinksAddressesByDevice(
                self._org_id,
                total_pages="all",
            )
        except meraki.exceptions.APIError as e:
            _LOGGER.debug("Organisation ID: %r: Get device statuses: %r", self._org_id, e)
            return []


class MerakiGetOrganizationApplianceUplinkStatuses(MerakiSectionOrg):
    @property
    def name(self):
        return f'getOrganizationApplianceUplinkStatuses_{self._org_id}'

    def get_live_data(self):
        try:
            return self._config.dashboard.appliance.getOrganizationApplianceUplinkStatuses(
                self._org_id, total_pages="all"
            )
        except meraki.exceptions.APIError as e:
            _LOGGER.debug("Organisation ID: %r: Get Appliance uplink status by network: %r", self._org_id, e)
            return []


class MerakiGetOrganizationApplianceUplinksUsageByNetwork(MerakiSectionOrg):
    @property
    def name(self):
        return f'getOrganizationApplianceUplinksUsageByNetwork_{self._org_id}'

    def get_live_data(self):
        if meraki.__version__ < "1.39.0":
            _LOGGER.debug(f'Meraki SDK is to old. Installed: {meraki.__version__}, expceted: 1.39.0' )
            return []
        try:
            return self._config.dashboard.appliance.getOrganizationApplianceUplinksUsageByNetwork(
                organizationId=self._org_id,
                total_pages="all",
                timespan=60  # default=86400 (one day), maximum=1209600 (14 days)
            )
        except meraki.exceptions.APIError as e:
            _LOGGER.debug("Organisation ID: %r: Get Appliance uplink usage by network: %r", self._org_id, e)
            return []


class MerakiGetOrganizationApplianceVpnStatuses(MerakiSectionOrg):
    @property
    def name(self):
        return f'getOrganizationApplianceVpnStatuses_{self._org_id}'

    def get_live_data(self):
        try:
            return self._config.dashboard.appliance.getOrganizationApplianceVpnStatuses(
                self._org_id,
                total_pages="all",
            )

        except meraki.exceptions.APIError as e:
            _LOGGER.debug("Organisation ID: %r: Get Appliance VPN status by network: %r", self._org_id, e)
            return []


class MerakiGetOrganizationSensorReadingsLatest(MerakiSectionOrg):
    @property
    def name(self):
        return f'getOrganizationSensorReadingsLatest_{self._org_id}'

    def get_live_data(self):
        try:
            return self._config.dashboard.sensor.getOrganizationSensorReadingsLatest(
                self._org_id, total_pages="all"
            )
        except meraki.exceptions.APIError as e:
            _LOGGER.debug("Organisation ID: %r: Get sensor readings: %r", self._org_id, e)
            return []


class MerakiGetDeviceSwitchPortsStatuses(MerakiSectionSerial):
    @property
    def name(self):
        return f'getDeviceSwitchPortsStatuses_{self._serial}'

    def get_live_data(self):
        try:
            return self._config.dashboard.switch.getDeviceSwitchPortsStatuses(
                self._serial, total_pages="all"
            )
        except meraki.exceptions.APIError as e:
            _LOGGER.debug("Serial: %r: Get Switch Port Statuses: %r", self._serial, e)
            return []


class MerakiGetOrganizationWirelessDevicesEthernetStatuses(MerakiSectionOrg):
    @property
    def name(self):
        return f'getOrganizationWirelessDevicesEthernetStatuses_{self._org_id}'

    def get_live_data(self):
        if meraki.__version__ < "1.39.0":
            _LOGGER.debug(f'Meraki SDK is to old. Installed: {meraki.__version__}, expceted: 1.39.0')
            return []
        try:
            return self._config.dashboard.wireless.getOrganizationWirelessDevicesEthernetStatuses(
                self._org_id, total_pages="all"
            )
        except meraki.exceptions.APIError as e:
            _LOGGER.debug("Organisation ID: %r: Get wireless devices ethernet statuses: %r", self._org_id, e)
            return []


class MerakiGetDeviceWirelessStatus(MerakiSectionSerial):
    @property
    def name(self):
        return f'getDeviceWirelessStatus_{self._serial}'

    def get_live_data(self):
        try:
            return self._config.dashboard.wireless.getDeviceWirelessStatus(
                self._serial
            )
        except meraki.exceptions.APIError as e:
            _LOGGER.debug("Serial: %r: Get wireless device status: %r", self._serial, e)
            return []


class MerakiGetOrganizationCellularGatewayUplinkStatuses(MerakiSectionOrg):
    @property
    def name(self):
        return f'getOrganizationCellularGatewayUplinkStatuses_{self._org_id}'

    def get_live_data(self):
        try:
            return self._config.dashboard.cellularGateway.getOrganizationCellularGatewayUplinkStatuses(
                self._org_id, total_pages="all"
            )
        except meraki.exceptions.APIError as e:
            _LOGGER.debug("Organisation ID: %r: Get cellular gateways uplink statuses: %r", self._org_id, e)
            return []


#
# Main run
#
@dataclass()
class MerakiOrganisation:
    config: MerakiConfig
    organisation: _Organisation
    # _piggyback_prefix = ''
    # _piggyback_suffix = ''
    # _piggyback_case = None
    _networks: dict[str, MerakiNetwork] | None = None

    @property
    def organisation_id(self) -> str:
        return self.organisation[_API_NAME_ORGANISATION_ID]

    @property
    def organisation_name(self) -> str:
        return self.organisation[_API_NAME_ORGANISATION_NAME]

    def query(self) -> Iterator[Section]:
        if organisation := MerakiGetOrganization(
                config=self.config,
                org_id=self.organisation_id).get_data(use_cache=self.config.use_cache):
            yield self._make_section(
                name=_SEC_NAME_ORGANISATIONS,
                data=organisation,
            )
            if not organisation[_API_NAME_API][_API_NAME_ENABLED]:
                # stop here if API is not enabled for this organisation
                return

        if _SEC_NAME_LICENSES_OVERVIEW not in self.config.excluded_sections:
            if licenses_overview := self._get_licenses_overview():
                yield self._make_section(
                    name=_SEC_NAME_LICENSES_OVERVIEW,
                    data=licenses_overview,
                )

        if networks := MerakiGetOrganizationNetworks(
            config=self.config,
            org_id=self.organisation_id
        ).get_data(use_cache=self.config.use_cache):
            yield from self._add_networks(networks)

        if _need_devices(self.config.excluded_sections):
            devices_by_serial = self._get_devices_by_serial()
        else:
            devices_by_serial = {}

        # stop here if there are no devices in the organisation
        if not devices_by_serial:
            return

        devices_by_type = {}
        for _device in devices_by_serial.values():
            if not devices_by_type.get(_device[_API_NAME_DEVICE_PRODUCT_TYPE]):
                devices_by_type[_device[_API_NAME_DEVICE_PRODUCT_TYPE]] = []
            devices_by_type[_device[_API_NAME_DEVICE_PRODUCT_TYPE]].append(_device)

        for device in devices_by_serial.values():
            try:
                device_piggyback = device[_API_NAME_DEVICE_NAME]
            except KeyError as e:
                _LOGGER.debug(
                    "Organisation ID: %r: Get device piggyback: %r", self.organisation_id, e
                )
                continue

            if not device_piggyback:  # skip devices without name, could use serial instead
                device_piggyback = f'{device[_API_NAME_DEVICE_SERIAL]}-{device[_API_NAME_DEVICE_PRODUCT_TYPE]}'
            yield self._make_section(
                name=_SEC_NAME_DEVICE_INFO,
                data=device,
                piggyback=self._adjust_piggyback(device_piggyback),
            )

        if _SEC_NAME_DEVICE_STATUSES not in self.config.excluded_sections:
            for device_status in MerakiGetOrganizationDevicesStatuses(
                    config=self.config,
                    org_id=self.organisation_id,
                    cache_interval=_MIN_CACHE_INTERVAL + randrange(_RANDOM_CACHE_INTERVAL),
            ).get_data(use_cache=self.config.use_cache):
                if piggyback := self._get_device_piggyback(device_status, devices_by_serial):
                    yield self._make_section(
                        name=_SEC_NAME_DEVICE_STATUSES,
                        data=device_status,
                        piggyback=self._adjust_piggyback(host=piggyback),
                    )
        if _SEC_NAME_DEVICE_UPLINKS_INFO not in self.config.excluded_sections:
            for device_uplink in MerakiGetOrganizationDevicesUplinksAddressesByDevice(
                    config=self.config,
                    org_id=self.organisation_id,
                    cache_interval=_MIN_CACHE_INTERVAL + randrange(_RANDOM_CACHE_INTERVAL),
            ).get_data(use_cache=self.config.use_cache):
                if piggyback := self._get_device_piggyback(device_uplink, devices_by_serial):
                    yield self._make_section(
                        name=_SEC_NAME_DEVICE_UPLINKS_INFO,
                        data=device_uplink,
                        piggyback=self._adjust_piggyback(host=piggyback),
                    )

        if devices_by_type.get(_API_NAME_DEVICE_TYPE_SENSOR):
            if _SEC_NAME_SENSOR_READINGS not in self.config.excluded_sections:
                for sensor_reading in MerakiGetOrganizationSensorReadingsLatest(
                        config=self.config,
                        org_id=self.organisation_id,
                        cache_interval=_MIN_CACHE_INTERVAL + randrange(_RANDOM_CACHE_INTERVAL)
                ).get_data(use_cache=self.config.use_cache):
                    if piggyback := self._get_device_piggyback(sensor_reading, devices_by_serial):
                        yield self._make_section(
                            name=_SEC_NAME_SENSOR_READINGS,
                            data=sensor_reading,
                            piggyback=self._adjust_piggyback(host=piggyback),
                        )

        if devices_by_type.get(_API_NAME_DEVICE_TYPE_APPLIANCE):
            if _SEC_NAME_APPLIANCE_UPLINKS not in self.config.excluded_sections:
                uplink_usage_by_network = MerakiGetOrganizationApplianceUplinksUsageByNetwork(
                        config=self.config,
                        org_id=self.organisation_id,
                        cache_interval=59,
                ).get_data()  # use_cache=self.config.use_cache her we want always life data

                # convert usage by network to usage by serial
                usage_by_serial = {}
                for network in uplink_usage_by_network:
                    for uplink in network['byUplink']:
                        usage_by_serial[uplink[_API_NAME_DEVICE_SERIAL]] = usage_by_serial.get(uplink[_API_NAME_DEVICE_SERIAL], {})
                        usage_by_serial[uplink[_API_NAME_DEVICE_SERIAL]].update(
                            {uplink['interface']: {'sent': uplink['sent'], 'received': uplink['received']}}
                        )

                for appliance_uplinks in MerakiGetOrganizationApplianceUplinkStatuses(
                        config=self.config,
                        org_id=self.organisation_id,
                        cache_interval=_MIN_CACHE_INTERVAL + randrange(_RANDOM_CACHE_INTERVAL)
                ).get_data(use_cache=self.config.use_cache):
                    if piggyback := self._get_device_piggyback(appliance_uplinks, devices_by_serial):
                        # add network name
                        appliance_uplinks['networkName'] = self._networks[appliance_uplinks[_API_NAME_NETWORK_ID]].name
                        # add uplink usage
                        if appliance_usage := usage_by_serial.get(appliance_uplinks[_API_NAME_DEVICE_SERIAL], None):
                            for uplink in appliance_uplinks['uplinks']:
                                uplink.update(appliance_usage.get(uplink['interface'], {}))

                        yield self._make_section(
                            name=_SEC_NAME_APPLIANCE_UPLINKS,
                            data=appliance_uplinks,
                            piggyback=self._adjust_piggyback(host=piggyback),
                        )

            if _SEC_NAME_APPLIANCE_VPNS not in self.config.excluded_sections:
                for appliance_vpn in MerakiGetOrganizationApplianceVpnStatuses(
                        config=self.config,
                        org_id=self.organisation_id,
                        cache_interval=_MIN_CACHE_INTERVAL + randrange(_RANDOM_CACHE_INTERVAL),
                ).get_data(use_cache=self.config.use_cache):
                    appliance_vpn.update({_API_NAME_DEVICE_SERIAL: appliance_vpn['deviceSerial']})  # devices_by_serial expects serial
                    if piggyback := self._get_device_piggyback(appliance_vpn, devices_by_serial):
                        yield self._make_section(
                            name=_SEC_NAME_APPLIANCE_VPNS,
                            data=appliance_vpn,
                            piggyback=self._adjust_piggyback(host=piggyback),
                        )

        if devices_by_type.get(_API_NAME_DEVICE_TYPE_SWITCH):
            if _SEC_NAME_SWITCH_PORTS_STATUSES not in self.config.excluded_sections:
                for switch in devices_by_type[_API_NAME_DEVICE_TYPE_SWITCH]:
                    ports_statuses = MerakiGetDeviceSwitchPortsStatuses(
                        config=self.config,
                        serial=switch[_API_NAME_DEVICE_SERIAL]
                    ).get_data(use_cache=self.config.use_cache)
                    if piggyback := self._get_device_piggyback(
                            {_API_NAME_DEVICE_SERIAL: switch[_API_NAME_DEVICE_SERIAL]}, devices_by_serial
                    ):
                        yield self._make_section(
                            name=_SEC_NAME_SWITCH_PORTS_STATUSES,
                            data=ports_statuses,
                            piggyback=self._adjust_piggyback(host=piggyback),
                        )

        if devices_by_type.get(_API_NAME_DEVICE_TYPE_WIRELESS):
            if _SEC_NAME_WIRELESS_ETHERNET_STATUSES not in self.config.excluded_sections:
                for device in MerakiGetOrganizationWirelessDevicesEthernetStatuses(
                        config=self.config,
                        org_id=self.organisation_id,
                        cache_interval=_MIN_CACHE_INTERVAL + randrange(_RANDOM_CACHE_INTERVAL),
                ).get_data(use_cache=self.config.use_cache):
                    if piggyback := self._get_device_piggyback(device, devices_by_serial):
                        yield self._make_section(
                            name=_SEC_NAME_WIRELESS_ETHERNET_STATUSES,
                            data=device,
                            piggyback=self._adjust_piggyback(host=piggyback),
                        )

            if _SEC_NAME_WIRELESS_DEVICE_STATUS not in self.config.excluded_sections:
                for device in devices_by_type[_API_NAME_DEVICE_TYPE_WIRELESS]:
                    wireless_statuses = MerakiGetDeviceWirelessStatus(
                        config=self.config,
                        serial=device[_API_NAME_DEVICE_SERIAL],
                    ).get_data(use_cache=self.config.use_cache)
                    if piggyback := self._get_device_piggyback({
                        _API_NAME_DEVICE_SERIAL: device[_API_NAME_DEVICE_SERIAL]
                    }, devices_by_serial):
                        yield self._make_section(
                            name=_SEC_NAME_WIRELESS_DEVICE_STATUS,
                            data=wireless_statuses,
                            piggyback=self._adjust_piggyback(host=piggyback),
                            )

        if devices_by_type.get(_API_NAME_DEVICE_TYPE_CELLULAR):
            if _SEC_NAME_CELLULAR_UPLINKS not in self.config.excluded_sections:
                for gateway in MerakiGetOrganizationCellularGatewayUplinkStatuses(
                        config=self.config,
                        org_id=self.organisation_id,
                        cache_interval=_MIN_CACHE_INTERVAL + randrange(_RANDOM_CACHE_INTERVAL)
                ).get_data(use_cache=self.config.use_cache):
                    if piggyback := self._get_device_piggyback(gateway, devices_by_serial):
                        yield self._make_section(
                            name=_SEC_NAME_CELLULAR_UPLINKS,
                            data=gateway,
                            piggyback=self._adjust_piggyback(host=piggyback),
                        )

    def _add_networks(self, networks):
        self._networks = {_API_NAME_ORGANISATION_NAME: self.organisation_name}
        for network in networks:
            network.update({'organizationName': self.organisation_name})
            self._networks[network['id']] = MerakiNetwork(
                id=network['id'],
                name=network['name'],
                product_types=network['productTypes'],
                time_zone=network['timeZone'],
                tags=network['tags'],
                enrollment_string=network['enrollmentString'],
                notes=network['notes'],
                is_bound_to_config_template=network['isBoundToConfigTemplate'],
                organisation_id=network['organizationId'],
                organisation_name=self.organisation_name,
                url=network['url'],
            )
        yield self._make_section(
            name=_SEC_NAME_NETWORKS,
            data=networks,
        )

    def _adjust_piggyback(self, host: str) -> str:
        if self.config.org_id_as_prefix:
            return f'{self.organisation_id}-{host}'
        return host

    def _get_licenses_overview(self) -> MerakiAPIData | None:
        def _update_licenses_overview(
            licenses_overview: dict[str, object] | None
        ) -> MerakiAPIData | None:
            if not licenses_overview:
                return None
            licenses_overview.update(
                {
                    "organisation_id": self.organisation_id,
                    "organisation_name": self.organisation_name,
                }
            )
            return licenses_overview

        return _update_licenses_overview(
            MerakiGetOrganizationLicensesOverview(
                config=self.config,
                org_id=self.organisation_id
            ).get_data(use_cache=self.config.use_cache)
        )

    def _get_devices_by_serial(self) -> Mapping[str, MerakiAPIData]:
        def _update_device(device: dict[str, object]) -> MerakiAPIData:
            device.update(
                {
                    "organisation_id": self.organisation_id,
                    "organisation_name": self.organisation_name,
                    'network_name': self._networks.get(device.get(_API_NAME_NETWORK_ID)).name,
                }
            )
            return device

        return {
            str(device[_API_NAME_DEVICE_SERIAL]): _update_device(device)
            for device in MerakiGetOrganizationDevices(
                config=self.config,
                org_id=self.organisation_id
            ).get_data(use_cache=self.config.use_cache)
        }

    def _get_device_piggyback(
        self, device: MerakiAPIData, devices_by_serial: Mapping[str, MerakiAPIData]
    ) -> str | None:
        try:
            serial = device[_API_NAME_DEVICE_SERIAL]
            if devices_by_serial[serial][_API_NAME_DEVICE_NAME]:
                return devices_by_serial[serial][_API_NAME_DEVICE_NAME]
            else:
                _LOGGER.debug(f'Host without name _get_device_piggyback serial: {serial}')
                return None
        except KeyError as e:
            _LOGGER.debug("Organisation ID: %r: Get device piggyback: %r", self.organisation_id, e)
            return None

    @staticmethod
    def _make_section(*, name: str, data: MerakiAPIData, piggyback: str | None = None) -> Section:
        return Section(
            api_data_source=MerakiAPIDataSource.org,
            name=_SECTION_NAME_MAP[name],
            data=data,
            piggyback=piggyback,
        )


def _query_meraki_objects(
        *,
        organisations: Sequence[MerakiOrganisation],
) -> Iterable[Section]:
    for organisation in organisations:
        yield from organisation.query()


def _write_sections(sections: Iterable[Section]) -> None:
    sections_by_piggyback: dict = {}
    for section in sections:
        sections_by_piggyback.setdefault(section.piggyback, {}).setdefault(
            section.get_name(), []
        ).append(section.data)

    for piggyback, pb_section in sections_by_piggyback.items():
        with ConditionalPiggybackSection(piggyback):
            for section_name, section_data in pb_section.items():
                with SectionWriter(section_name) as writer:
                    writer.append_json(section_data)


# .
#   .--main----------------------------------------------------------------.
#   |                                       _                              |
#   |                       _ __ ___   __ _(_)_ __                         |
#   |                      | '_ ` _ \ / _` | | '_ \                        |
#   |                      | | | | | | (_| | | | | |                       |
#   |                      |_| |_| |_|\__,_|_|_| |_|                       |
#   |                                                                      |
#   '----------------------------------------------------------------------'


def parse_arguments(argv: Sequence[str] | None) -> argparse.Namespace:
    parser = create_default_argument_parser(description=__doc__)

    parser.add_argument("hostname")
    parser.add_argument(
        "apikey",
        help="API key for the Meraki API dashboard access.",
    )

    parser.add_argument("--proxy", type=str)

    parser.add_argument(
        "--sections",
        nargs="+",
        choices=list(_SECTION_NAME_MAP),
        default=list(_SECTION_NAME_MAP),
        help="Explicit sections that are collected.",
    )

    parser.add_argument(
        "--excluded-sections",
        nargs="+",
        choices=list(_SECTION_NAME_MAP),
        default=[],
        help="Sections that are excluded form data collected.",
    )

    parser.add_argument(
        "--orgs",
        nargs="+",
        default=[],
        help="Explicit organisation IDs that are checked.",
    )

    parser.add_argument(
        "--prefix-suffix",
        nargs=5,
        action='append',
        default=[],
        help="Explicit organisation IDs that are checked.",
        metavar=('Org-by-ID/Name', 'Organisation', 'Change Case', 'Prefix', 'Suffix')
    )

    parser.add_argument(
        '--no-cache',
        default=False,
        action='store_const',
        const=True,
        help='Never use cached information. By default the agent will cache received data to '
             'avoid API limits and speed up the data retrieving.'
    )

    parser.add_argument(
        '--org-id-as-prefix',
        default=False,
        action='store_const',
        const=True,
        help='Use organisation ID as hostname prefix.'
    )

    return parser.parse_args(argv)


def _need_devices(section_names: Sequence[str]) -> bool:
    return any(
        s not in section_names
        for s in [
            _SEC_NAME_APPLIANCE_UPLINKS,
            _SEC_NAME_APPLIANCE_VPNS,
            _SEC_NAME_CELLULAR_UPLINKS,
            _SEC_NAME_DEVICE_STATUSES,
            _SEC_NAME_DEVICE_UPLINKS_INFO,
            _SEC_NAME_SENSOR_READINGS,
            _SEC_NAME_SWITCH_PORTS_STATUSES,
            _SEC_NAME_WIRELESS_DEVICE_STATUS,
            _SEC_NAME_WIRELESS_ETHERNET_STATUSES,
        ]
    )


def _get_organisations(config: MerakiConfig, org_ids: Sequence[str]) -> Sequence[_Organisation]:
    organisations = [
        _Organisation(
            id=organisation[_API_NAME_ORGANISATION_ID],
            name=organisation[_API_NAME_ORGANISATION_NAME],
        ) for organisation in MerakiGetOrganizations(
            config=config
        ).get_data(use_cache=config.use_cache)
    ]
    if org_ids:
        organisations = [
            organisation for organisation in organisations if organisation[_API_NAME_ORGANISATION_ID] in org_ids
        ]
    return organisations


def agent_cisco_meraki_main(args: Args) -> int:
    # don't remove used for runtime logging at the end
    start_time = time_ns()
    config = MerakiConfig(
        dashboard=_configure_meraki_dashboard(
            args.apikey,
            args.debug,
            args.proxy,
        ),
        hostname=args.hostname,
        # section_names=args.sections,
        excluded_sections=args.excluded_sections,
        use_cache=not args.no_cache,
        org_id_as_prefix=args.org_id_as_prefix,
    )

    organisations = [
        MerakiOrganisation(config, organisation)
        for organisation in _get_organisations(config, args.orgs)
    ]

    sections = _query_meraki_objects(
        organisations=organisations,
    )

    _write_sections(sections)

    _LOGGER.warning(f'Time taken: {(time_ns() - start_time) / 1e9}/s')
    _LOGGER.warning(f'Meraki SDK version: {meraki.__version__}')
    return 0


def main() -> int:
    return special_agent_main(parse_arguments, agent_cisco_meraki_main)
