#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright (C) 2022 Checkmk GmbH - License: GNU General Public License v2
# This file is part of Checkmk (https://checkmk.com). It is subject to the terms and
# conditions defined in the file COPYING, which is part of this source code package.

# enhancements by thl-cmk[at]outlook[dot]com, https://thl-cmk.hopto.org
# - changed check_last_reported_ts to output report as metric/levels
# - added levels_upper check_last_reported_ts
# - added section names from the cisco meraki special agent (for use in WATO)

import json
import time
from collections.abc import Mapping, Sequence
from typing import Final
from dataclasses import dataclass

from cmk.base.plugins.agent_based.agent_based_api.v1 import render, Result, State, check_levels
from cmk.base.plugins.agent_based.agent_based_api.v1.type_defs import CheckResult, StringTable

MerakiAPIData = Mapping[str, object]

_SEC_NAME_ORGANISATIONS: Final = "_organisations"  # internal use runs always
_SEC_NAME_DEVICE_INFO: Final = "_device_info"  # Not configurable, needed for piggyback
_SEC_NAME_NETWORKS: Final = "_networks"  # internal use, runs always, needed for network names

_SEC_NAME_LICENSES_OVERVIEW: Final = "licenses-overview"
_SEC_NAME_DEVICE_STATUSES: Final = "device-statuses"
_SEC_NAME_SENSOR_READINGS: Final = "sensor-readings"
_SEC_NAME_DEVICE_UPLINKS_INFO: Final = "device-uplinks-info"
_SEC_NAME_APPLIANCE_UPLINKS: Final = "appliance-uplinks"
_SEC_NAME_APPLIANCE_VPNS: Final = "appliance-vpns"
_SEC_NAME_SWITCH_PORTS_STATUSES: Final = "switch-ports-statuses"
_SEC_NAME_WIRELESS_ETHERNET_STATUSES: Final = "wireless-ethernet-statuses"
_SEC_NAME_WIRELESS_DEVICE_STATUS: Final = "wireless-device-status"
_SEC_NAME_CELLULAR_UPLINKS: Final = "cellular-uplinks"


@dataclass(frozen=True)
class MerakiNetwork:
    id: str  # "N_24329156",
    name: str  # "Main Office",
    product_types: Sequence[str]  # ["appliance", "switch", "wireless"]
    time_zone: str  # "America/Los_Angeles",
    tags: Sequence[str]  # [ "tag1", "tag2" ],
    enrollment_string: str | None  # "my-enrollment-string",
    notes: str  # "Additional description of the network",
    is_bound_to_config_template: bool  # false
    organisation_id: str
    organisation_name: str
    url: str


def load_json(string_table: StringTable) -> Sequence[MerakiAPIData]:
    try:
        return json.loads(string_table[0][0])
    except (IndexError, json.decoder.JSONDecodeError):
        return []


def check_last_reported_ts(
        last_reported_ts: float,
        levels_upper: Sequence[int, int] | None = None,
        as_metric: bool | None = True) -> CheckResult:

    if (age := time.time() - last_reported_ts) < 0:
        yield Result(
            state=State.OK,
            summary="Negative timespan since last report time.",
        )
        return
    if levels_upper:
        levels_upper = (levels_upper[0] * 3600, levels_upper[1] * 3600)

    yield from check_levels(
        value=age,
        label='Time since last report',
        metric_name='last_reported' if as_metric else None,
        levels_upper=levels_upper,
        render_func=render.timespan,
    )
