#!/usr/bin/env python3
# Copyright (C) 2019 Checkmk GmbH - License: GNU General Public License v2
# This file is part of Checkmk (https://checkmk.com). It is subject to the terms and
# conditions defined in the file COPYING, which is part of this source code package.

# enhancements by thl-cmk[at]outlook[dot]com, https://thl-cmk.hopto.org
# - changed remaining time (WATO) from Age (Days, Hours, Minutes, Seconds) to Days only
# - added WATO option for License state is not ok -> default to WARN
# - added discovery rule for ITEM variant (Org Name/Org ID - this is the default, Org Name, Org ID)

# 2023-11-18: moved discovery rule set to cisco_meraki_organisations for reuse with cisco_meraki_organisations_api

from cmk.gui.i18n import _
from cmk.gui.plugins.wato.utils import (
    CheckParameterRulespecWithItem,
    rulespec_registry,
    RulespecGroupCheckParametersApplications,
)
from cmk.gui.valuespec import (
    Integer,
    Dictionary,
    TextInput,
    Tuple,
    MonitoringState,
    FixedValue,
)


def _migrate(p: str | dict) -> dict[str, any]:
    # change age to days, we suspect there is no warning less than 1 hour in seconds,
    # and we will not warn/crit above 3600 days (10 years)
    if p.get('remaining_expiration_time'):
        warn, crit = p.get('remaining_expiration_time')
        if warn >= 3600:
            warn = int(warn / 86400)
        if crit >= 3600:
            crit = int(crit / 86400)
        p['remaining_expiration_time'] = (warn, crit)
    return p


def _parameter_valuespec_cisco_meraki_org_licenses_overview():
    return Dictionary(
        title=_("Cisco Meraki Organisation Licenses Overview"),
        optional_keys=True,
        elements=[
            ("remaining_expiration_time",
             Tuple(
                 title=_("Lower levels for remaining expiration time of licenses"),
                 elements=[
                     Integer(title=_("Warning at"), unit='Days'),
                     Integer(title=_("Critical at"), unit='Days'),
                 ],
             )),
            ("state_license_not_ok",
             MonitoringState(
                 title=_('Monitoring state if License state is not OK'),
                 default_value=1,
             )),
            ('dont_show_alias_on_info',
             FixedValue(
                 value=True,
                 title=_('Don\'t show alias on info line'),
                 totext='',
                 help=_(
                     'The alias is the Organisation ID or the Organisation name, depending on the Item.\n'
                     'If the item is the Organisation ID, the alias is the Organisation name and vice versa.\n'
                     'Organisation ID and Organisation name will always shown in the service details.'
                 )
             ))
        ],
        migrate=_migrate,
        ignored_keys=[
            'internal_item_name',
            'old_item_name',
            'item_variant',
        ],
    )


rulespec_registry.register(
    CheckParameterRulespecWithItem(
        title=lambda: _("Cisco Meraki Organisation Licenses Overview"),
        check_group_name="cisco_meraki_org_licenses_overview",
        group=RulespecGroupCheckParametersApplications,
        parameter_valuespec=_parameter_valuespec_cisco_meraki_org_licenses_overview,
        item_spec=lambda: TextInput(
            title=_("The organisation"),
        ),
        match_type="dict",
    )
)
